/***************************************************************
Copyright (C) 2005  Yannet Interian and Gabriel Corvera

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
******************************************************************/ 

#include "qbf.h"

extern struct Formula F;
extern struct Subformula SubF;



FILE * CutCommentsFile(char *name)
{
  FILE *fp;
  int lastc;
  int nextc;


  if ((fp = fopen(name,"r"))==NULL) {
    printf("No such graph as %s\n",name);
    exit(1);
  }

  while ((lastc =  fgetc(fp)) == 'c')
      {
	//read until you get and End of Line (We are consuming the comments of the file)
	while ((nextc = fgetc(fp)) != EOF && nextc != '\n');
      }
    ungetc(lastc,fp);
	//get the first line of the problem. It must be a CNF formula
    if (fscanf(fp,"p cnf %i %i",&F.numatom,&F.numclause) != 2)
      {
	fprintf(stderr,"Bad input file\n");
	exit(-1);
      }

    if(F.numatom > MAXATOM)
      {
	fprintf(stderr,"ERROR - too many atoms\n");
	exit(-1);
      }
    if(F.numclause > MAXCLAUSE)
      {
	fprintf(stderr,"ERROR - too many clauses\n");
	exit(-1);
      }

	//return the file pointer without p line nor comments
  return fp;
}


void processFile(FILE *fp, char *name)
{
  int i,j,lit;
  int *storeptr1,*storeptr2;
  int freestore1=0,freestore2=0;
  int size,sizeU,sizeE;
  int tmp[MAXLENGTH];

  //VARIABLES
  for (i=1;i<=F.numatom;i++){
    	F.atoms[i]=0;
  }
  F.numatomU=0;
  F.numatomE=0;


  //universal variables are read here
  fscanf(fp," a  %i ",&j);
  //printf(" %i ",j);
  while(j!=0){
    F.numatomU++;
    F.atomU[F.numatomU]=j;
    F.atoms[j]=1;
    fscanf(fp,"%i ",&j);

    //printf(" %i ",j);

  }
  printf("\n");
  //existential variables are read here
  fscanf(fp," e  %i ",&j);
  //printf(" %i ",j);
  while(j!=0){
    F.numatomE++;
    F.atomE[F.numatomE]=j;
    F.atoms[j]=2;
    fscanf(fp,"%i ",&j);

    //printf(" %i ",j);
  }
  //printf("\n");

  for(i = 0;i < 2*MAXATOM+1;i++){
    F.numoccurence[i] = 0;
  }


  //FORMULA
  printf(" numclause %d \n",F.numclause);
  for(i = 0;i < F.numclause;i++){
    	size=0;
    	do{
		if (fscanf(fp,"%i ",&lit) != 1){
	  	fprintf(stderr, "Bad input file i:%d\n",i);
	  	exit(-1);
		}

		//if the literal is not the terminal (0) we fill the temporal Array
		if(lit != 0){
		  tmp[size]=lit;
		  size++;
		  F.numoccurence[lit+MAXATOM]++;
		  // printf("%d ",lit);
		}
	    }while(lit != 0);

    	if(size > MAXLENGTH){
      		printf("ERROR - clause too long\n");
      		exit(-1);
    	}
    //printf("\n");

    sizeU=0;
    sizeE=0;
    for (j=0;j<size;j++){

     if (freestore1 < MAXLENGTH){
       storeptr1 = (int *) malloc( sizeof(int) * STOREBLOCK );
       freestore1 = STOREBLOCK;
       fprintf(stderr,"allocating memory...\n");
     }
     if (freestore2 < MAXLENGTH){
       storeptr2 = (int *) malloc( sizeof(int) * STOREBLOCK );
       freestore2 = STOREBLOCK;
       fprintf(stderr,"allocating memory...\n");
     }


     if (F.atoms[abs(tmp[j])]==1){
       if (sizeU==0){
	 F.clauseU[i] = storeptr1;
       }
       sizeU++;
       *(storeptr1++) = tmp[j];
       freestore1--;
     }

     if (F.atoms[abs(tmp[j])]==2){
       if (sizeE==0)
	 F.clauseE[i] = storeptr2;
       sizeE++;
       *(storeptr2++) = tmp[j];
       freestore2--;
     }

    }
    F.sizeU[i]=sizeU;
    F.sizeE[i]=sizeE;

  }
  fclose(fp);
}



void TwoaltQBFinput( char *name){
  FILE *fp;
  fp=CutCommentsFile(name);
  processFile(fp,name);
  //printingFormula();
  occurrences();
  //printingOccurrences();

}


void occurrences(){
  int *storeptr;
  int freestore=0;
  int i,j,aux;
  for(i = 0;i < 2*MAXATOM+1;i++){
    if (freestore < F.numoccurence[i])
      {
		storeptr = (int *) malloc( sizeof(int) * STOREBLOCK );
		freestore = STOREBLOCK;
		fprintf(stderr,"allocating memory...\n");
      }
    F.occurence[i] = storeptr;
    freestore -= F.numoccurence[i];
    storeptr += F.numoccurence[i];
    F.numoccurence[i] = 0;
  }

  for(i = 0;i < F.numclause;i++){
    for(j = 0;j < F.sizeU[i];j++){
	aux=F.clauseU[i][j];
	F.occurence[aux+MAXATOM]
	  [F.numoccurence[aux+MAXATOM]] = i;
	F.numoccurence[aux+MAXATOM]++;
      }
    for(j = 0;j < F.sizeE[i];j++){
	aux=F.clauseE[i][j];
	F.occurence[aux+MAXATOM]
	  [F.numoccurence[aux+MAXATOM]] = i;
	F.numoccurence[aux+MAXATOM]++;
    }
  }

}


void printingOccurrences()
{
  int i,j,aux;
  for(i = 0;i < 2*MAXATOM+1;i++){
    if (F.numoccurence[i]>0){
      aux=i-MAXATOM;
      printf("lit %d ",aux);
      for (j=0;j<F.numoccurence[i];j++)
	printf(" %d ",F.occurence[i][j]);
      printf("\n");
    }
  }
}

void printingFormula(){
  int i,j;
  printingAtoms();
  for (i=0;i<F.numclause;i++){
    for (j=0;j<F.sizeU[i];j++){
      printf(" %3d ",F.clauseU[i][j]);
    }
    for (j=0;j<F.sizeE[i];j++){
      printf(" %3d ",F.clauseE[i][j]);
    }
    printf("\n");
  }

}



void printingAtoms(){
  int i;
  printf("printing atoms \n");
  for (i=1;i<=F.numatom;i++)
    printf(" %d ", F.atoms[i]);
  printf("\n");
}


