/***************************************************************
Copyright (C) 2005  Yannet Interian and Gabriel Corvera

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
******************************************************************/

// remember ...   ./walkQBF -h  for info

#include "qbf.h"

int subformulas = 0;
static struct option long_options[] = {
  {"file", 1, 0, 'f'},
  {0, 0, 0, 0}
};

extern void print_sol2(int);

static int scratch;
#define ABS(x) ((scratch=(x))>0?(scratch):(-scratch))

/************************************/
/* for the original formula             */

/************************************/

char* inputfile;


int STEPS=0;
int seed=1;
long int idum=(-1) ;

struct Formula F;
struct Subformula SubF;
int atom[MAXATOM+1];


struct SOL S[3];




/************************************/
/* Forward declarations             */
/************************************/

/**
 * Alarm Implementation and variables
 */

 int kill_flag = FALSE; /*finish flag*/
 int run_time = 60*60*2; /* Total Running Time*/

 void kill_program(int signum){
 	kill_flag = TRUE;
 }

/************************************/
/* Main                             */
/************************************/

int main (int argc,char * argv[])
{
  int i,t;
  int aux;
  long begin;
  clock_t Start, Finish;
  struct tms *strtemps;
  double UserTime =0;

   struct timeval tv;
   struct timezone tzp;

   gettimeofday(&tv,&tzp);


  /** Start TIME*/
  begin = tv.tv_sec;
  strtemps = ( struct tms *) malloc( sizeof (struct tms));
  t = times(strtemps);
  Start = strtemps->tms_utime;

  parsecommandline(argc,argv);

  signal(SIGALRM, kill_program);
  /* Set alarm clock for the time assigned*/
  alarm(run_time);

  TwoaltQBFinput(inputfile);

  WITHwalkminsat(STEPS);  //second approach


  i = times(strtemps);
  Finish = strtemps->tms_utime;
  UserTime = (double) (Finish-Start)/(double) CLK_TCK;
  printf("Process Time : %.3f seconds\n", UserTime);

  gettimeofday(&tv,&tzp);
  printf("Total Time: %i seconds\n", tv.tv_sec-begin);
  free(strtemps);

  return 0;
}



void WITHwalkminsat()
{
  int i;
  int aux,aux2;
  int steps=1000000000;
  int numsubform=0;


  for (i=1;i<steps;i++){
    runminsat();
    assignment2subformula_1(0);
    numsubform++;
    if (call_kcnfs(numsubform))
      break;
    // assignment2subformula_1(1);
    //numsubform++;
    //if (call_kcnfs(numsubform))
    //break;
    assignment2subformula_1(2);
    numsubform++;
    if (call_kcnfs(numsubform))
      break;
    if(kill_flag){
    	printf("\n\tTimeout\n");
    	break;
    }
  }
}






/*given an assigment to the universals
  calculates which clauses are in the subformula*/
void assignment2subformula()
{
  int i,j;
  int aux;
  SubF.numclause=0;
  for (i=0;i<F.numclause;i++){
    SubF.clause[i]=1;
    for (j=0;j<F.sizeU[i];j++){
      aux=F.clauseU[i][j];
      if ((aux > 0) == atom[ABS(aux)]){
        SubF.clause[i]=0;
        break;
      }
    }
  }
}


void assignment2subformula_1(int k)
{
  int i,j;
  int aux;
  SubF.numclause=F.numclause;
  for (i=0;i<F.numclause;i++){
    SubF.clause[i]=1;
    for (j=0;j<F.sizeU[i];j++){
      aux=F.clauseU[i][j];
      if ((aux > 0) == S[k].solution[ABS(aux)]){
		SubF.numclause--;
        SubF.clause[i]=0;
        break;
      }
    }
  }
}

void computeDegrees()
{
  int i,j,aux;
  //calulating degrees corresponding to sub
  for (i=1;i<=F.numatomE;i++){
    aux=F.atomE[i];
    SubF.numocc[MAXATOM+aux]=0;
    SubF.numocc[MAXATOM-aux]=0;
  }

  for (i=0;i<F.numclause;i++){
    if (SubF.clause[i]==1){
      SubF.numclause++;
      for (j=0;j<F.sizeE[i];j++){
        aux=F.clauseE[i][j];
        //printf(" %d ",aux);
        SubF.numocc[MAXATOM+aux]++;
      }
      //printf("\n");
    }
  }
}



// measure on how balance is a subformula
int balanceSub()
{
  int aux,aux1,aux2,i,auxscore=0;
  int dscore=0;
  for (i=1;i<=F.numatomE;i++){
    aux=F.atomE[i];
    aux1=SubF.numocc[MAXATOM+aux];
    aux2=SubF.numocc[MAXATOM-aux];
    auxscore+=abs(aux1-aux2);
  }
  dscore=auxscore;
  return dscore;
}



int call_kcnfs(int numsubform)
{
  int isUnsat=0;

  /*printf("--------sub-formula %d \n",numsubform);*/
    ExitSubformula(numsubform);
    isUnsat=mainKCNFS(numsubform);
//    fprintf(stderr,".");
    if(isUnsat==20){
      printf("\n\tUNSAT was found\n\n");
      printf("\tUNSAT SUBFORMULA: %i\n\n",numsubform);
      return 1; /*if the subformula is unsat we stop looking for sub-formulas */
    }
    return 0;
}




/*================Printing =================== */

void PrintAssignmentU()
{
  int i;
  for (i=1;i<=F.numatomU;i++){
    printf(" %d ",atom[F.atomU[i]]);
  }
  printf("\n");
}



void PrintSubformula()
{
  int i,aux,aux1,aux2;
  for (i=0;i<F.numclause;i++){
    printf(" %d ",SubF.clause[i]);
  }
  printf("\n");
  for (i=1;i<=F.numatomE;i++){
    aux=F.atomE[i];
    aux1=SubF.numocc[MAXATOM+aux];
    aux2=SubF.numocc[MAXATOM-aux];
    printf("%3d: %3d    -%3d: %3d \n",aux,aux1,aux,aux2);
  }

}


/*=============== End Printing ===================*/



/*===============      Exits    =================*/

void ExitSubformula(int a)
{
  max_length = 0, min_length = KMAXLENGTH;
  _b_leaders = OFF, _b_breakfix = OFF, cnf2 = OFF;
  _ssat = NULL;
  int i,j,aux;
  char  name[100];

#ifdef DEBUG_SUBFORMULAS_OUTPUT
  FILE* fp;
#endif

  boolean _b;
  int k, _c = 0, myc[KMAXLENGTH+1], _myc, ki,kj;
  char s[100], c;
  struct cnf *_cnf;
  struct wcl *_wcl;

  _nb_clause = nb_clause;

#ifdef DEBUG_SUBFORMULAS_OUTPUT
  snprintf(name, 100, "subformula-%d.cnf",a);
  fp=fopen(name,"w");
#endif

  //Begin to output the subformula
#ifdef DEBUG_SUBFORMULAS_OUTPUT
  fprintf(fp,"c this formula was genrerated by walkQBF \n");
  fprintf(fp,"p cnf %d %d \n",F.numatomE+F.numatomU,SubF.numclause);
#endif

  //Number of variables and clauses of the subformula
  nb_var = F.numatomE+F.numatomU; //Change made as suggested by Gilles to make KCNFS compatible con all formulas
   nb_clause = SubF.numclause;

  //We are going to generate the subformular verifying the contents of the
  //other clauses.
  for (i=0;i<F.numclause;i++){ //EQUIVALENT TO THE FOR IN LINE 863 OF KCNFS.C
    //If a subformula is found
    _myc = 0;
    if (SubF.clause[i]==1){ //THE STRUCTURES OF KCNF MUST ONLY FILL WITHIN THIS
      //Copy the clauses of the subformula to the datastructure CNF
      for (j=0;j<F.sizeE[i];j++){
	aux=F.clauseE[i][j];
#ifdef DEBUG_SUBFORMULAS_OUTPUT
	fprintf(fp,"%d ",aux);
#endif
	for(_b = OK, ki = 0; ki < _myc; ki++)
	  //verify the same integer is not in the myc array
	  if(myc[ki] == aux) { _b = FAULT; break; }
	//if everything is ok continue
	if(_b) {
	  myc[_myc++] = aux;
	  if(_myc > KMAXLENGTH+1) {
	    printf("\nc readcnf :: CLAUSE SIZE CANNOT EXCEED %d !!", KMAXLENGTH);
	    return;
	  }
	}

      }

#ifdef DEBUG_SUBFORMULAS_OUTPUT
      fprintf(fp,"0 \n");
#endif
      //Add the 0 to the array of variables so KCNFS can work correctly
      myc[_myc++] = 0;

      _b = OK;
      for (ki = 0; (_c = myc[ki]); ki++) {
	for (kj = 0; (aux = myc[kj]); kj++)
	  if (_c == -aux) { --nb_clause; _b = FAULT; break; }
	if (!_b) break;
      }
      if(_b) {
	if(_myc > max_length) max_length = _myc;
	if(_myc < min_length) min_length = _myc;
	_cnf = (struct cnf* )built_clause(myc);
	_cnf->next = _ssat;
	_ssat = _cnf;
      }

    }
  }
  for (i = 1; i <= nb_var; i++) {
    _wcl_p[i] = _wcl_n[i] = NULL;
    v_icu[i] = v_icu_l[i + nb_var] = v_icu_l[-i + nb_var] = 0;
  }
  _cnf = _ssat;
  while(_cnf != NULL) {
    for(i = 0; (j = _cnf->c[i]); i++) {
      _wcl = (struct wcl*)built_addr(_cnf);
      if(j > 0) { _wcl->next = _wcl_p[j]; _wcl_p[j] = _wcl; }
      else { _wcl->next = _wcl_n[-j]; _wcl_n[-j] = _wcl; }
    }
    _cnf = _cnf->next;
  }
  if(--min_length < 3) {
    cnf2 = ON;
    min_length = 3;
  }

  init_max_length = max_length;
  --max_length;

#ifdef DEBUG_SUBFORMULAS_OUTPUT
  fclose(fp);
#endif

}
/*=============================================*/





int parsecommandline(int argc, char **argv)
//get command line parameters
{
        char c;
        char h[20];
		heuristicParam=-1;
        while((c=getopt_long(argc, argv, "m:f:N:s:t:h",long_options,NULL))!=-1) {
                switch (c) {
                case 'f':
                  inputfile=optarg;
                  break;
                case 'N':
                  STEPS=atoi(optarg);
                  break;
                case 's':
                  seed=atoi(optarg);
                  idum*=seed;
                  break;
                case 't':
                	run_time=atoi(optarg);
                	break;
                case 'm':
                    snprintf(h,20,"%s",optarg); 
                	if(strcmp(h,"MIXED_BALANCE")==0)
	                	heuristicParam=MIXED_BALANCE;
                	else if(strcmp(h,"RANDOM")==0)
	                	heuristicParam=RANDOM;
                	else if(strcmp(h,"WORST")==0)
	                	heuristicParam=WORST;
                	else if(strcmp(h,"MIXED")==0)
	                	heuristicParam=MIXED;
                	else if(strcmp(h,"BALANCE")==0)
	                	heuristicParam=BALANCE;	                		                		                	
                	else if(strcmp(h,"MIXED_BLOCK")==0)
	                	heuristicParam=MIXED_BLOCK;
                	else if(strcmp(h,"BEST_AND_SECOND")==0)
	                	heuristicParam=BEST_AND_SECOND;
	                else{
	                	printf("The given heuristic doesn't exist\n");
	                	exit(-1);
	                }	                	               		                
                	break;
                case 'h':
                default:
                        fprintf(stderr, "%s [options]\n"
                                "\n"
                                       "\t-f or --file <input file >\n"
                                "\t-t <time in seconds>\n"
                                "\t-m <HEURISTIC>\n"
                                "\t-h  this help\n"
                                "\nAvailable Heuristics: RANDOM, WORST, BALANCE, MIXED, MIXED_BLOCK, BEST_AND_SECOND, MIXED_BALANCE\n"
                                ,argv[0]);
                        exit(-1);
                }
        }
        return 0;
}


