/***************************************************************
Copyright (C) 2005  Yannet Interian and Gabriel Corvera

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
******************************************************************/ 

/* originaly written on top of walksat */

#define VERSION 1

/************************************/
/* Compilation flags                */
/************************************/

/***************************************************************/
/* Please select only one of following flags: UNIX, ANSI or NT */
/* Description:                                                */
/*    UNIX:  uses some non-POSIX unix time routines, compile   */
/*           successful under SunOS 5.6, time accuracy 1/60sec */
/*    ANIS:  use POSIX time routines, can compile under all    */
/*           unix and NT, time accuracy is 1sec                */ 
/*    NT:    use NT/Win32 multimedia routines, compile under   */
/*           NT only, time accuracy is 1ms                     */
/*           Uses NT "timeGetTime" function:                   */
/*             Header: Decl. in Mmsystem.h; include Windows.h. */
/*             Library: Use Winmm.lib.                         */
/***************************************************************/

#define UNIX 1
/* #define ANSI 1 */
/* #define NT 1 */


/* Define BIGINT to be the type used for the "cutoff" variable.
   Under gcc "long long int" gives a 64 bit integer.
   Program will still function using a 32-bit value, but it
   limit size of cutoffs that can be specified. */

#ifdef UNIX
#define BIGINT long long int
#endif
#ifdef NT
#define BIGINT __int64
#endif
#ifdef ANSI
#define BIGINT long int
#endif


/************************************/
/* Standard includes                */
/************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/types.h>
#include <limits.h>
#include <signal.h>
#include <assert.h>
#include <stdarg.h>


#include "qbf.h"


#ifdef UNIX
  #include <sys/times.h>
  #include <sys/time.h>
#endif
#ifdef NT
  #include <time.h>
  #include <windows.h>
  #include <mmsystem.h>
#endif
#ifdef ANSI
  #include <sys/time.h>
#endif

#ifndef UNIX   /* ANSI or NT */
  #define random() rand()
  #define srandom(seed) srand(seed)
#endif

/* Sometimes CLK_TCK is defined in a weird place */
#ifndef CLK_TCK
  #define CLK_TCK 60
#endif



/************************************/
/* Constant parameters              */
/************************************/


//#define MAXATOM 150000		/* maximum possible number of atoms */


#define TRUE 1
#define FALSE 0

//#define MAXLENGTH 500           /* maximum number of literals which can be in any clause */

/************************************/
/* Internal constants               */
/************************************/

#define NOVALUE -1


#define Var(CLAUSE, POSITION) (ABS(clauseU[CLAUSE][POSITION]))

static int scratch;
#define ABS(x) ((scratch=(x))>0?(scratch):(-scratch))

#define BIG 100000000

/************************************/
/* Main data structures             */
/************************************/

/* Atoms start at 1 */
/* Not a is recorded as -1 * a */

/**
 * Definition of Data Structures and Variables Needed to solve
 * MinSAT
 */

int numtruelit[MAXCLAUSE];	/* number of true literals in each clause */
int bestminsat;			/* value used to know the best minsat number of true clauses */
int wheretrue[MAXCLAUSE]; /* where each clause is listed in true */
int true[MAXCLAUSE]; /* clauses which are true */	
int whereonetrue[MAXCLAUSE]; /* where each cluase is listed in onetrue */
int onetrue[MAXCLAUSE]; /* clauses which have only on true literal */
int numtrue;			/* number of true clauses */
int numonetrue; 		/* number of only one true literal clauses */

extern int atom[MAXATOM+1];		/* value of each atom */ 
	
extern struct Formula F;
extern struct Subformula SubF;



extern struct SOL S[3];


int breakcount[MAXATOM+1];	/* number of clauses that become unsat if var if flipped */
int makecount[MAXATOM+1];	/* number of clauses that become sat if var if flipped */



int TheTrueOne[MAXCLAUSE];


/**************************************
 * Data Structure used for mixed_block
 *************************************/

int blocked[MAXATOM];
int blockprob=10;
int balancenoise=10;  //=20;

/************************************/
/* Global flags and parameters      */
/************************************/

int status_flag = 0;		/* value returned from main procedure */
int abort_flag;
int make_flag;
int balance_flag;

int heuristic = WORST;

int numerator = NOVALUE;	/* make random flip with numerator/denominator frequency */
int denominator = 100;		

BIGINT numflip;                /* number of changes so far*/
int numrun = 10;
BIGINT cutoff = 100000;


int numtry = 0;			/* total attempts at solutions */



/* Randomization */

unsigned int seed;  /* Sometimes defined as an unsigned long int */

#ifdef UNIX
   struct timeval tv;
   struct timezone tzp;
#endif
#ifdef NT
   DWORD win_time;     /* elapsed time in ms, since windows boot up */
#endif


/************************************/
/* Forward declarations             */
/************************************/

void parse_parameters(int argc,char *argv[]);
void print_parameters(int argc, char * argv[]);
void init_minsat(void);

int pickrandom(void);
int pickworst(void);
int pickbalance(void);
int pickmixed(void);
int pickmixed_2(void);
int pickmixed_block(void);
int pickbest_and_second(void);
int pickmixed_balance(void);

char * heuristic_names[] = { "random","worst","balance","mixed","mixed_2","mixed_block","best_and_second","mixed_balance"};

static int (*pickcode[])(void) =
      {pickrandom, pickworst,pickbalance,pickmixed,pickmixed_2,pickmixed_block,pickbest_and_second,pickmixed_balance};


double elapsed_seconds(void);
void scanone(int argc, char *argv[], int i, int *varptr);
void scanonell(int argc, char *argv[], int i, BIGINT *varptr);
void init();
void flipatom(int toflip);

void save_solution(int);

void print_current_assign(void);
void handle_interrupt(int sig);

long super(int i);

void print_sol_file(char * filename,int);
void print_sol_cnf(int);
void print_sol(int);
void print_sol2(int);


void init_balance();
void flipatom_balance(int toflip);
int d_balance(int y,int flag);
int undo_balance(int y);




/************************************/
/* Main                             */
/************************************/


int runminsat()
{

  int aux;
  double aux2;

  int bestscore=BIG;
  int bestsum=BIG;

  init_minsat();

#ifdef UNIX
    gettimeofday(&tv,&tzp);
    seed = (unsigned int)((( tv.tv_sec & 0177 ) * 1000000) + tv.tv_usec);
#endif
#ifdef NT
    seed = (unsigned int)(timeGetTime());
#endif
#ifdef ANSI
    seed = (unsigned int)(time());
#endif

    
    parse_parameters(0, NULL);
    srandom(seed);
    /*    print_parameters(argc, argv); */
    
    (void) elapsed_seconds();

    while (! abort_flag && numtry < numrun) {
      numtry++;
      init();

      numflip=0;
      while(numflip < cutoff) {	
	numflip++;
	//flipatom((pickcode[heuristic])());
	
		flipatom_balance((pickcode[heuristic])());
		
		if(numtrue<bestminsat){
		  save_solution(0);
		  bestminsat = numtrue;
		}	
	
		
		if ((SubF.score/2+numtrue)<bestsum){
		  save_solution(2);
		  bestsum = SubF.score/2+numtrue;
		}
      }
    }
    //expertime = elapsed_seconds();
    
    
 /*   aux=F.numclause-bestminsat;
    aux2=((float)aux)/F.numatomE;
   printf("\n Minimal Solution Found: (%d true clauses): (%d false clauses) \n",bestminsat,aux);
    printf(" ratio %.2f \n",aux2);
    print_sol(0);
	print_sol2(0);
    print_sol(2);
    print_sol2(2);
    printf("\n");   */
    return status_flag;
}



void init_minsat(){

  if(heuristicParam>=0){
  	 heuristic = heuristicParam;
  }else{
	  heuristic = MIXED_BALANCE;
  }
  status_flag = 0;		/* value returned from main procedure */
  
  numerator = 0;	/* make random flip with numerator/denominator frequency */
  denominator = 100;		

  numrun = 1; //I change this  august 26, Yannet
  numtry=0;
  cutoff = 2000; //100000
  bestminsat=BIG;
}



void parse_parameters(int argc,char *argv[])
{
    int i;
    int temp;

    for (i=1;i < argc;i++)
    {
	if (strcmp(argv[i],"-seed") == 0)
	    scanone(argc,argv,++i,&seed);
	//else if (strcmp(argv[i],"-out") == 0 && i<argc-1)
	//  strcpy(outfile, argv[++i]);
	else if (strcmp(argv[i],"-status") == 0)
	    status_flag = 1;
	else if (strcmp(argv[i],"-cutoff") == 0)
	    scanonell(argc,argv,++i,&cutoff);
	else if (strcmp(argv[i],"-random") == 0)
	    heuristic = RANDOM;
	else if (strcmp(argv[i],"-noise") == 0){
	    scanone(argc,argv,++i,&numerator);
	    if (i < argc-1 && sscanf(argv[i+1],"%i",&temp)==1){
		denominator = temp;
		i++;
	    }
	}
        
	else if (strcmp(argv[i],"-tries") == 0 || strcmp(argv[i],"-restart") == 0)
	    scanone(argc,argv,++i,&numrun);
	
	else 
	{
	    fprintf(stderr, "General parameters:\n");
	    fprintf(stderr, "  -seed N -cutoff N -restart N\n");
	    fprintf(stderr, "Heuristics:\n");
	    fprintf(stderr, "  -random -worst \n");
	    fprintf(stderr, "  -noise N or -noise N M (default M = 100)\n");
	    exit(-1);
	}
    }
    //base_cutoff = cutoff;
    
    if (numerator==NOVALUE){
	switch(heuristic) {
	case WORST:
	  numerator = 50;
	    break;
	default:
	    numerator = 0;
	    break;
	}
    }
}



void print_parameters(int argc, char * argv[])
{
  int i;

  printf("%i\n", VERSION);
  printf("command line =");
  for (i=0;i < argc;i++){
    printf(" %s", argv[i]);
  }
  printf("\n");
  printf("seed = %u\n",seed);
  printf("cutoff = %Li\n",cutoff);
  printf("tries = %i\n",numrun);

  printf("heuristic = ");
  switch(heuristic)
    {
    default:
      printf("%s", heuristic_names[heuristic]);
      break;
    }
  if (numerator>0){
    printf(", noise %d / %d", numerator, denominator);
  }
  
  printf("\n");
}






void scanone(int argc, char *argv[], int i, int *varptr)
{
    if (i>=argc || sscanf(argv[i],"%i",varptr)!=1){
	fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
	exit(-1);
    }
}

void scanoneu(int argc, char *argv[], int i, unsigned int *varptr)
{
    if (i>=argc || sscanf(argv[i],"%u",varptr)!=1){
	fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
	exit(-1);
    }
}

void scanonell(int argc, char *argv[], int i, BIGINT *varptr)
{
    if (i>=argc || sscanf(argv[i],"%Li",varptr)!=1){
	fprintf(stderr, "Bad argument %s\n", i<argc ? argv[i] : argv[argc-1]);
	exit(-1);
    }
}




void init()
{
  int i;
  int j;
  int thetruelit;


  for(i = 0;i < F.numclause;i++){
    numtruelit[i] = 0;
  }
  numtrue = 0;
  numonetrue = 0;
  /* Random Initialization of Atom */
  for(i = 1;i < F.numatomU+1;i++){
    atom[F.atomU[i]] = random()%2;  
    
    //breakcount[i] = 0;
    makecount[F.atomU[i]] = 0;
    /*For Blocked Mix Heuristic*/
    if(random()%blockprob==0){
    	blocked[F.atomU[i]]=1;
    }else{
    	blocked[F.atomU[i]]=0;    	
    }
  }
  init_balance();

  /* Initialize breakcount and makecount in the following: */
  for(i = 0;i < F.numclause;i++)
    {
      for(j = 0;j < F.sizeU[i];j++)
	{
	  if((F.clauseU[i][j] > 0) == atom[ABS(F.clauseU[i][j])]){ 
	    numtruelit[i]++;
	    thetruelit = F.clauseU[i][j];
	  }
	}
      if(numtruelit[i] == 0)
	{
	  for(j = 0;j < F.sizeU[i];j++){
	    makecount[ABS(F.clauseU[i][j])]++;
	  }
	}
      else{
	if(numtruelit[i] > 0)
	  {
	    true[numtrue] = i;
	    wheretrue[i] = numtrue;
	    numtrue++;
	    //change this
	    TheTrueOne[i]=thetruelit;  //to assign a true literal even to non onetrue clauses
	  }
	if (numtruelit[i] == 1)
	  {
	    //TheTrueOne[i]=thetruelit;
	    //breakcount[ABS(thetruelit)]++;
	    onetrue[numonetrue] = i;
	    whereonetrue[i] = numonetrue;
	    numonetrue++; 
	  }
      }
    }
}
	



void flipatom(int toflip)
{
  int i,j;		
  int toenforce;		
  register int cli;
  register int lit;
  int numocc;
  register int sz;
  register int * litptr;
  int * occptr;

  if(atom[toflip] > 0){
    toenforce = -toflip;
  }
  else{
    toenforce = toflip;
  }
  
  /* flipping the atom */
  atom[toflip] = 1-atom[toflip];
  

  numocc = F.numoccurence[MAXATOM+toenforce];
  occptr = F.occurence[MAXATOM+toenforce];

  
  for(i = 0; i < numocc; i++)
    {
      /* cli = occurence[MAXATOM+toenforce][i]; */
      cli = *(occptr++);
      
      if (++numtruelit[cli] == 1){
	/* add to one true and true since now it is a true clause */
	TheTrueOne[cli]=toenforce;
	onetrue[numonetrue] = cli;
	whereonetrue[cli] = numonetrue;
	numonetrue++;
	true[numtrue] = cli;
	wheretrue[cli] = numtrue;
	numtrue++;
	/* Decrement the makecount of all vars in the clause */
	sz = F.sizeU[cli];
	litptr = F.clauseU[cli];
	for (j=0; j<sz; j++){
	  /* lit = clause[cli][j]; */
	  lit = *(litptr++);
	  makecount[ABS(lit)]--;
	}
      }
      else if (numtruelit[cli] == 2){
	/* remove from one true */   
	numonetrue--;
	onetrue[whereonetrue[cli]]=onetrue[numonetrue]; 
	whereonetrue[onetrue[numonetrue]]=whereonetrue[cli];
	
      }
    }
  
  numocc = F.numoccurence[MAXATOM-toenforce];
  occptr = F.occurence[MAXATOM-toenforce];

  
  for(i = 0; i < numocc ;i++)
    {
      /* cli = occurence[MAXATOM-toenforce][i]; */
      cli = *(occptr++);
      
      if (--numtruelit[cli] == 0){
	/* It is now a false clause, erase it from one and onetrue */
	numonetrue--;
	numtrue--;
	true[wheretrue[cli]] = true[numtrue];
	wheretrue[true[numtrue]] = wheretrue[cli];
	onetrue[whereonetrue[cli]] = onetrue[numonetrue];
	whereonetrue[onetrue[numonetrue]] = whereonetrue[cli];		  
	/* Increment the makecount of all vars in the clause */
	sz = F.sizeU[cli];
	litptr = F.clauseU[cli];
	for (j=0; j<sz; j++){
	  /* lit = clause[cli][j]; */
	  lit = *(litptr++);
	  makecount[ABS(lit)]++;
	}
      }
      else if (numtruelit[cli] == 1){
	/* now it is a one true clause */
	sz = F.sizeU[cli];
	litptr = F.clauseU[cli];
	for (j=0; j<sz; j++){
	  lit = *(litptr++);
	  if ((lit > 0) == atom[ABS(lit)]){
	    TheTrueOne[cli]=lit;
	    break;
	  }
	}
	onetrue[numonetrue] = cli;
	whereonetrue[cli] = numonetrue;
	numonetrue++;   
      }
    }
  
  
}
	
/***************Heuristics********************/



int pickmixed()
{

  int num=10;
  if (random()%denominator < num)
    return pickbalance();
  else
    return pickworst();
  
}



int pickbalance()
{

  int i,number=7,dscore;
  int tofix[MAXLENGTH];
  int best[MAXLENGTH];
  register int numbest=0;
  register int bestvalue=BIG;
  register int var;

  if(numonetrue==0){
    return pickrandom();
  }

  for (i=0;i<number;i++)
    tofix[i] = onetrue[random()%numonetrue];
  for (i=0;i<number;i++){
    var = ABS(TheTrueOne[tofix[i]]);
    dscore = d_balance(var,0);
    d_balance(var,1);
    if (dscore<=bestvalue){
      if (dscore<bestvalue)
	numbest=0;
      bestvalue = dscore;
      best[numbest++] = var;
    }
  }
  if (bestvalue>0 && (random()%denominator < numerator))
    return ABS(TheTrueOne[tofix[0]]);
  
  if (numbest == 1) 
    return best[0];
  return best[random()%numbest];
  
}



/* choose a random clauses in onetrue and flip the sat lit */
int pickrandom(void){

  int tofix;
  
  if(numonetrue>0){
    tofix = onetrue[random()%numonetrue];			
  }else if(numtrue>0){
    tofix = true[random()%numtrue];	
  }else{
    assert(0); /* must not enter here */
  }

  //change this for the case numonetrue=0
  return ABS(TheTrueOne[tofix]);
}



/* choose form two random clauses in onetrue the literal that satisies less clauses  */
int pickworst(void){

  int i,number=10,nummake;
  int tofix[MAXLENGTH];
  int best[MAXLENGTH];
  register int numbest;
  register int bestvalue;
  register int var;

  numbest = 0;
  bestvalue = BIG;
  
  if (numtrue==0){
    printf("ERROR in pickworst \n");
    return -1;
  }

  if(numonetrue>0){
    for (i=0;i<number;i++)
      tofix[i] = onetrue[random()%numonetrue];
  }
  else if(numtrue>0){
    printf(" numonetrue=0 \n");
    for (i=0;i<number;i++)
      tofix[i] = true[random()%numtrue];	
  }

  //change this ... for the case numonetrue=0
  for (i=0;i<number;i++){
    var = ABS(TheTrueOne[tofix[i]]);
    nummake = makecount[var];
    if (nummake<=bestvalue){
      if (nummake<bestvalue)
	numbest=0;
      bestvalue = nummake;
      best[numbest++] = var;
    }
  }
  if (bestvalue>0 && (random()%denominator < numerator))
    return ABS(TheTrueOne[tofix[0]]);
  
  if (numbest == 1) 
    return best[0];
  return best[random()%numbest];
}


int pickmixed_2(void)
{
  int i,number=10,nummake;
  int tofix[MAXLENGTH];
  int best[MAXLENGTH];
  int best2=0;
  register int numbest;
  register int bestvalue;
  register int var;
  register int dscore;

  numbest = 0;
  bestvalue = BIG;
  
  if (numtrue==0){
    printf("ERROR in pickworst \n");
    return -1;
  }

  if(numonetrue>0){
    for (i=0;i<number;i++)
      tofix[i] = onetrue[random()%numonetrue];
  }
  else if(numtrue>0){
    printf(" numonetrue=0 \n");
    for (i=0;i<number;i++)
      tofix[i] = true[random()%numtrue];	
  }

  //change this ... for the case numonetrue=0
  for (i=0;i<number;i++){
    var = ABS(TheTrueOne[tofix[i]]);
    nummake = makecount[var];
    if (nummake<=bestvalue){
      if (nummake<bestvalue)
		numbest=0;
      bestvalue = nummake;
      best[numbest++] = var;
    }
  }
  if (bestvalue>0 && (random()%denominator < numerator))
    return ABS(TheTrueOne[tofix[0]]);

  
  if (numbest == 1) 
    return best[0];

  bestvalue = BIG;

  for (i=0;i<numbest;i++){
    var = best[i];

    dscore = d_balance(var,0);
    d_balance(var,1);
    
    if (dscore<bestvalue){
      bestvalue = dscore;
      best2 = var;
    }
  }
  return best2;
}


int pickmixed_block(void)
{
  int i,number=10,nummake,jumped=0;
  int tofix[MAXLENGTH];
  int best[MAXLENGTH];
  int best2=0;
  register int numbest;
  register int bestvalue;
  register int var;
  register int dscore;
	
	
  numbest = 0;
  bestvalue = BIG;
  
  if (numtrue==0){
    printf("ERROR in pickworst \n");
    return -1;
  }

  if(numonetrue>0){
    for (i=0;i<number;i++)
      tofix[i] = onetrue[random()%numonetrue];
  }
  else if(numtrue>0){
    printf(" numonetrue=0 \n");
    for (i=0;i<number;i++)
      tofix[i] = true[random()%numtrue];	
  }

  //change this ... for the case numonetrue=0
  for (i=0;i<number;i++){
    var = ABS(TheTrueOne[tofix[i]]);
    if(blocked[var]==1){
    	jumped++;
    	continue;
    }
    nummake = makecount[var];
    if (nummake<=bestvalue){
      if (nummake<bestvalue)
		numbest=0;
      bestvalue = nummake;
      best[numbest++] = var;
    }
  }
  if(jumped==number){
	  printf("All Clauses were jumped\n");
  }
  if (bestvalue>0 && (random()%denominator < numerator))
    return ABS(TheTrueOne[tofix[0]]);

  
  if (numbest == 1) 
    return best[0];

  bestvalue = BIG;

  for (i=0;i<numbest;i++){
    var = best[i];
    dscore = d_balance(var,0);
    d_balance(var,1);
    if (dscore<bestvalue){
      bestvalue = dscore;
      best2 = var;
    }
  }

  return best2;
}


int pickbest_and_second(){

	int best[MAXATOM];
	int second[MAXATOM];
  	int tofix[MAXLENGTH];	
	int bestnum=BIG;
	int secondnum=BIG;
	int i;
	int *a=best;
	int *b=second;
	int *temp;
	int nummake;
	int number=5;
	int best2=0;
	register int numa;
	register int numb;
  	int bestvalue;
  	int var;
	int dscore;

	numa=0;
	numb=0;
	bestvalue = BIG;

  	if (numtrue==0){
    	printf("ERROR in pickbest_and_second \n");
    	return -1;
  	}

  	if(numonetrue>0){
    	for (i=0;i<number;i++)
      		tofix[i] = onetrue[random()%numonetrue];
  	}
  	else if(numtrue>0){
    	printf(" numonetrue=0 \n");
    	for (i=0;i<number;i++)
      		tofix[i] = true[random()%numtrue];	
  	}


	for(i = 0; i < number; i++){
   	 	var = ABS(TheTrueOne[tofix[i]]);
    	nummake = makecount[var];				
		if (nummake<=bestnum){

			if(nummake<bestnum){
				temp = b;
				b = a;
				a = temp;
				numb = numa;
				numa = 0;
				secondnum=bestnum;
				bestnum=nummake;
			}
			a[numa++] = var;

		}else if(nummake<=secondnum){
			if(nummake<secondnum){
				numb = 0;
				secondnum = nummake;
			}
			b[numb++]=var;

		}
	}

	for(i = 0; i < numa; i++){
	    var = a[i];
	    dscore = d_balance(var,0);
	    d_balance(var,1);
	    if (dscore<bestvalue){	    	
	      bestvalue = dscore;
	      best2 = var;
	    }
	}
	for(i = 0; i < numb; i++){
		var = b[i];
	    dscore = d_balance(var,0);
	    d_balance(var,1);
	    if (dscore<bestvalue){
	      bestvalue = dscore;
	      best2 = var;
	    }
	}
	
	return best2;
}

int pickmixed_balance(void){

	if(random()%balancenoise==0){
		return pickbest_and_second();
	}else{
		return pickmixed_2();
	}
		
}

double elapsed_seconds(void) 
{ 
  double answer;
  
#ifdef ANSI
  static long prev_time = 0;
  time( &long_time );
  /* Note:  time(&t) returns t in seconds, so do not need to /CLK_TCK */
  answer = long_time - prev_time;
  prev_time = long_time;
#endif
#ifdef NT
  static DWORD prev_time = 0;
  win_time = timeGetTime();
  /* Note:  return value of timeGetTime() is ms, so divide by 1000*/
  answer = (double)(win_time - prev_time) / (double)1000; 
  prev_time = win_time;
#endif
#ifdef UNIX
  static struct tms prog_tms;
  static long prev_times = 0;
  (void) times(&prog_tms);
  answer = ((double)(((long)prog_tms.tms_utime)-prev_times))/((double) CLK_TCK);
  prev_times = (long) prog_tms.tms_utime;
#endif
  return answer; 
}




void print_sol_cnf(int k)
{
  int i;
  for(i = 1;i < F.numatomU+1;i++){
    printf("v %i\n", S[k].solution[F.atomU[i]] == 1 ? i : -i);
    
  }
}


void print_sol(int k)
{
  int i;
  
  for(i = 1;i < F.numatomU+1;i++){
    printf("%i ", S[k].solution[F.atomU[i]]);		
  }
  printf("\n");
  printf("score %d numtrue %d  mix score %d \n",S[k].score,S[k].numtrue,S[k].mix_score);

}

void print_sol2(int k)
{
  int i;
  
  for(i = 1;i < F.numatomU+1;i++){
    printf("%i ", S[k].solution[F.atomU[i]]>0?F.atomU[i]:-F.atomU[i]);		
  }
  printf("\n");
  printf("score %d numtrue %d  mix score %d \n",S[k].score,S[k].numtrue,S[k].mix_score);

}


void print_sol_file(char * filename,int k)
{
    FILE * fp;
    int i;

    if ((fp = fopen(filename, "w"))==NULL){
	fprintf(stderr, "Cannot open output file\n");
	exit(-1);
    }
    for(i = 1;i < F.numatomU+1;i++){
	fprintf(fp, " %i", S[k].solution[F.atomU[i]] == 1 ? i : -i);
	if (i % 10 == 0) fprintf(fp, "\n");
    }
    if ((i-1) % 10 != 0) fprintf(fp, "\n");
    fclose(fp);
}



void print_current_assign(void)
{
    int i;

    printf("Begin assign at flip = %d \n", numflip);
    for (i=1; i<=F.numatomU; i++){
	printf(" %d", atom[F.atomU[i]]==0 ? -i : i);
	if (i % 10 == 0) printf("\n");
    }
    if ((i-1) % 10 != 0) printf("\n");
    printf("End assign\n");
}





void save_solution(int k)
{
  int i;
  
  for (i=1; i<=F.numatomU; i++){
    S[k].solution[F.atomU[i]] = atom[F.atomU[i]];
  }
  S[k].score=SubF.score;
  S[k].numtrue=numtrue;
  S[k].mix_score=SubF.score/20 + numtrue;
}




/********************** for balance ****************/

void init_balance()
{

  assignment2subformula();
  computeDegrees();
  SubF.score=balanceSub();
  
}


//like flipatom but encoraging balance

void flipatom_balance(int toflip)
{
  int i,j;		
  int toenforce;		
  register int cli;
  register int lit;
  int numocc;
  register int sz;
  register int * litptr;
  int * occptr;

  if(atom[toflip] > 0){
    toenforce = -toflip;
  }
  else{
    toenforce = toflip;
  }
  
  /* flipping the atom */
  atom[toflip] = 1-atom[toflip];
  
  numocc = F.numoccurence[MAXATOM+toenforce];
  occptr = F.occurence[MAXATOM+toenforce];
  
  for(i = 0; i < numocc; i++)
    {
      /* cli = occurence[MAXATOM+toenforce][i]; */
      cli = *(occptr++);
      
      if (++numtruelit[cli] == 1){
	/* add to one true and true since now it is a true clause */
	TheTrueOne[cli]=toenforce;
	onetrue[numonetrue] = cli;
	whereonetrue[cli] = numonetrue;
	numonetrue++;
	true[numtrue] = cli;
	wheretrue[cli] = numtrue;
	numtrue++;
	/*change balance */
	sz = F.sizeE[cli];
	litptr = F.clauseE[cli];
	for (j=0; j<sz; j++){
	  /* lit = clause[cli][j]; */
	  lit = *(litptr++);
	  if (SubF.numocc[MAXATOM+lit]>=SubF.numocc[MAXATOM-lit]+1)
	    SubF.score--;
	  else
	    SubF.score++;
	  SubF.numocc[MAXATOM+lit]--;
	}
	
	/* Decrement the makecount of all vars in the clause*/ 
	sz = F.sizeU[cli];
	litptr = F.clauseU[cli];
	for (j=0; j<sz; j++){
	  /* lit = clause[cli][j]; */
	  lit = *(litptr++);
	  makecount[ABS(lit)]--;
	}
      
      }
      else if (numtruelit[cli] == 2){
	/* remove from one true */   
	numonetrue--;
	onetrue[whereonetrue[cli]]=onetrue[numonetrue]; 
	whereonetrue[onetrue[numonetrue]]=whereonetrue[cli];
	
      }
    }
  
  numocc = F.numoccurence[MAXATOM-toenforce];
  occptr = F.occurence[MAXATOM-toenforce];

  
  for(i = 0; i < numocc ;i++)
    {
      /* cli = occurence[MAXATOM-toenforce][i]; */
      cli = *(occptr++);
      
      if (--numtruelit[cli] == 0){
	/* It is now a false clause, erase it from one and onetrue */
	numonetrue--;
	numtrue--;
	true[wheretrue[cli]] = true[numtrue];
	wheretrue[true[numtrue]] = wheretrue[cli];
	onetrue[whereonetrue[cli]] = onetrue[numonetrue];
	whereonetrue[onetrue[numonetrue]] = whereonetrue[cli];		  

	sz = F.sizeE[cli];
	litptr = F.clauseE[cli];
	for (j=0; j<sz; j++){
	  /* lit = clause[cli][j]; */
	  lit = *(litptr++);
	  if (SubF.numocc[MAXATOM+lit]>SubF.numocc[MAXATOM-lit]-1)
	    SubF.score++;
	  else
	    SubF.score--;
	  SubF.numocc[MAXATOM+lit]++;
	}

	/* Increment the makecount of all vars in the clause */
	sz = F.sizeU[cli];
	litptr = F.clauseU[cli];
	for (j=0; j<sz; j++){
	  /* lit = clause[cli][j]; */
	  lit = *(litptr++);
	  makecount[ABS(lit)]++;
	}
      }
      else if (numtruelit[cli] == 1){
	/* now it is a one true clause */
	sz = F.sizeU[cli];
	litptr = F.clauseU[cli];
	for (j=0; j<sz; j++){
	lit = *(litptr++);
	if ((lit > 0) == atom[ABS(lit)]){
	  TheTrueOne[cli]=lit;
	    break;
	  }
	}
	onetrue[numonetrue] = cli;
	whereonetrue[cli] = numonetrue;
	numonetrue++;   
      }
    }
  
  
}
	



//differece in balance if variable y is flipped
int d_balance(int toflip,int flag)
{
  int i,j;		
  int toenforce;		
  register int cli;
  register int lit;
  int numocc;
  register int sz;
  register int * litptr;
  int * occptr;
  int dscore=0;
  
  if(atom[toflip] > 0){
    toenforce = -toflip;
  }
  else{
    toenforce = toflip;
  }
  
  
  numocc = F.numoccurence[MAXATOM+toenforce];
  occptr = F.occurence[MAXATOM+toenforce];
  
  for(i = 0; i < numocc; i++)
    {
      /* cli = occurence[MAXATOM+toenforce][i]; */
      cli = *(occptr++);
      
      if (numtruelit[cli] == 0){
		/* add to one true and true since now it is a true clause */
		
		/* balance */
		sz = F.sizeE[cli];
		litptr = F.clauseE[cli];
		for (j=0; j<sz; j++){
		  /* lit = clause[cli][j]; */
		  lit = *(litptr++);
		
		  if (SubF.numocc[MAXATOM+lit]>=SubF.numocc[MAXATOM-lit]+1){
		    dscore--;
		  }
		  else{
		    dscore++;
		  }
		  if(flag == 0){
			  SubF.numocc[MAXATOM+lit]--;
		  }else{
		  	  SubF.numocc[MAXATOM+lit]++;
		  }
		}
      }
    }
  
  numocc = F.numoccurence[MAXATOM-toenforce];
  occptr = F.occurence[MAXATOM-toenforce];

  for(i = 0; i < numocc ;i++)
    {
      /* cli = occurence[MAXATOM-toenforce][i]; */
      cli = *(occptr++);
      
      if (numtruelit[cli] == 1){
	
		sz = F.sizeE[cli];
		litptr = F.clauseE[cli];
		for (j=0; j<sz; j++){
		  /* lit = clause[cli][j]; */
		  lit = *(litptr++);
		
		  if (SubF.numocc[MAXATOM+lit]>SubF.numocc[MAXATOM-lit]-1){
		    dscore++;
		  }
		  else{
		    dscore--;
		  }
		  if(flag == 0){
			  SubF.numocc[MAXATOM+lit]++;
		  }
		  else{
  			  SubF.numocc[MAXATOM+lit]--;
		  }
		}
      }
    }

    
  return dscore;
}

/****************** End for balance ****************/
